import fs from "fs";
import TestUnit from "./TestUnit.mjs";

/**
 * Loader
 *
 * @package Tramline
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class Loader {

	/**
	 * Constructor.
	 *
	 * @param {string} root Test directory
	 * @param {string} tests optional Test file matcher (".test.mjs")
	 * @return {self}
	 */
	constructor({root=null, tests=".test.mjs"}={}) {

		/**
		 * Test directory
		 *
		 * @var {string}
		 */
		this.root = root;

		/**
		 * Test file matcher ("ends with")
		 *
		 * @return {string}
		 */
		this.tests = tests;

	}


	/**
	 * Load all tests.
	 *
	 * @return {array} Test units
	 */
	load() {
		let tests = [];
		this.files().forEach(file => {
			tests.push(new TestUnit(this.test(file), file));
		});
		return tests;
	}


	/**
	 * Find tests in a directory (recursive).
	 *
	 * @param {string} path optional Root directory (`this.root`)
	 * @return {array} Test files.
	 */
	files(root=this.root) {
		let files = [];
		fs.readdirSync(root).forEach(item => {
			let path = root + "/" + item;
			if (fs.statSync(path).isDirectory()) {
				files = files.concat(this.files(path));
			}
			else if (this.path(path)) {
				files.push(path);
			}
		});
		return files;
	}


	/**
	 * Get file path stripped of root.
	 *
	 * @param {string} path
	 * @return {bool}
	 */
	file(path) {
		return path.replace(this.root, "");
	}


	/**
	 * Get whether a file path is a valid test.
	 *
	 * Matches whether the path ends with the `this.tests` syntax.
	 *
	 * @param {string} path
	 * @return {bool}
	 */
	path(path) {
		return path.endsWith(this.tests);
	}


	/**
	 * Get class/test name from file path.
	 *
	 * @param {string} path
	 * @return {string}
	 */
	test(path) {
		let filePath = this.file(path).replace("\\", "/");
		let fileParts = filePath.split("/");
		let name = fileParts[fileParts.length - 1];
		let nameClass = name.replace(this.tests, "");
		return nameClass;
	}

}

export default Loader;