import TestHookError from "./Errors/TestHookError.mjs";

/**
 * ANSI codes
 *
 * @var {object}
 */
const ANSI = {
	reset: "\x1b[0m",
	color: {
		black: "\x1b[30m",
		red: "\x1b[31m",
		green: "\x1b[32m",
		yellow: "\x1b[33m",
		blue: "\x1b[34m",
		magenta: "\x1b[35m",
		cyan: "\x1b[36m",
		white: "\x1b[37m"
	},
	background: {
		black: "\x1b[40m",
		red: "\x1b[41m",
		green: "\x1b[42m",
		yellow: "\x1b[43m",
		blue: "\x1b[44m",
		magenta: "\x1b[45m",
		cyan: "\x1b[46m",
		white: "\x1b[47m"
	},
	bright: "\x1b[1m",
	dim: "\x1b[2m",
	underline: "\x1b[4m",
	blink: "\x1b[5m",
	reverse: "\x1b[7m",
	hidden: "\x1b[8m"
}

/**
 * ANSI boolean options
 *
 * @var {array}
 */
const ANSI_BOOLEAN = [
	"reset",
	"bright", "dim", "underline",
	"blink", "reverse", "hidden"
];

/**
 * Indentation multiplier
 *
 * @var {int}
 */
const INDENTATION = 1;

/**
 * Suite output strings
 *
 * @var {object}
 */
const SUITE = {
	"\n -> Test summary:\n": {
		bright: true
	},
	"PASS:\t{passed}": {
		indent: 1,
		bright: true,
		color: "green"
	},
	"FAIL:\t{failed}": {
		indent: 1,
		bright: true,
		color: "red"
	},
	"UNITS:\t{countUnits}": {
		indent: 1,
		bright: true
	},
	"TESTS:\t{count}":{
		indent: 1,
		bright: true
	},
	"\n {successFormatted}% OK after {timerFormatted}s.": {
		bright: true
	},
	"\n Tramline testing completed.": {
		bright: true
	}
};


/**
 * Logger
 *
 * @package Tramline
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class Logger {

	/**
	 * Constructor.
	 *
	 * @return {self}
	 */
	constructor() {

		/**
		 * Indent level
		 *
		 * @var {int}
		 */
		this.indent = 0;

		/**
		 * Default styles
		 *
		 * @var {object}
		 */
		this.styles = {
			"fail": ANSI.color.red + ANSI.bright,
			"pass": ANSI.color.green + ANSI.bright,
			"failHighlight": ANSI.color.black + ANSI.background.red,
			"passHighlight": ANSI.color.black + ANSI.background.green
		};

		/**
		 * Disable all output
		 *
		 * @var {bool}
		 */
		this.silent = false;

		/**
		 * Bind methods
		 */
		this.log = this.log.bind(this);
		this.error = this.error.bind(this);

	}


	/**
	 * Log a message.
	 *
	 * @param {mixed} message
	 * @param {object} optional options
	 * @return {void}
	 */
	log(message, options={}) {
		if (!this.silent) {
			console.log(this.stdout(message, options));
		}
	}


	/**
	 * Log an error.
	 *
	 * @param {mixed} error
	 * @return {void}
	 */
	error(error) {
		if (!this.silent) {
			if (error instanceof Error) {
				console.error(this.stdout(error.stack, {color: "red"}));
			}
			console.error(this.stdout(error, {color: "red"}));
		}
	}


	/**
	 * App header.
	 *
	 * @param {int} time
	 * @param {Config} config
	 * @param {array} tests
	 * @param {string} version
	 */
	app(time, config, tests, version) {
		let strings = [
			`\n -> Tramline v${version}`,
			` -> ${this.appTime(time)}`,
			` -> ${config.root}; ${tests.length} unit(s) found.`
		];
		this.separator();
		strings.forEach(string => this.log(string));
		this.separator();
	}


	/**
	 * Generate app time string for log.
	 *
	 * Uses `Date.toLocaleString(...)`.
	 *
	 * IMPORTANT: Locale is runtime-dependent, but `Europe/London` 
	 * 	timezone is currently hardcoded - TODO Make this config.
	 *
	 * @param {int} time
	 * @return {string}
	 */
	appTime(time) {
		return (new Date(time)).toLocaleString(undefined, {
			hour12: false,
			timeZone: "Europe/London",
			weekday: "long",
			month: "long",
			year: "numeric",
			day: "2-digit",
			hour: "2-digit",
			minute: "2-digit",
			second: "2-digit"
		});
	}


	/**
	 * Log a test after completion.
	 *
	 * @param {Test} test
	 * @return {string}
	 */
	test(test) {
		this.log(this.testMessage(test), {indent: 1});
		if (!test.passed) this.error(test.error);
	}


	/**
	 * Generate output message for a completed test.
	 *
	 * @param {Test} test
	 * @return {string}
	 */
	testMessage(test) {
		let msg;
		let status = test.passed ? "pass" : "fail";
		let styles = this.styles[status];
		let stylesStatus = this.styles[`${status}Highlight`];
		msg = `${stylesStatus} ${status.toUpperCase()} ${ANSI.reset}`;
		msg += `${styles} ${test.title}`;
		return msg;
	}


	/**
	 * Log a unit.
	 *
	 * @param {Unit} unit
	 * @return {void}
	 */
	unit(unit) {
		this.log("");
		this.log(` -> ${unit.name}\n`, {color: "yellow", bright: true});
		this.indent++;
	}


	/**
	 * End a unit.
	 *
	 * Resets identation.
	 *
	 * @return {void}
	 */
	unitEnd() {
		this.indent--;
	}


	/**
	 * Abort a unit.
	 *
	 * @param {mixed} error Error if any
	 * @return {void}
	 */
	unitAbort(error) {
		let msg = "Aborting test unit; see above output for details.";
		if (error) this.error(error);
		if (error instanceof TestHookError) {
			msg = "Critical test hook failed; aborting this test unit.";
		}
		this.log(` (!) ${msg}`, {color: "red", bright: true});
	}


	/**
	 * Log summary output for a suite.
	 *
	 * Output determined by `SUITE`; `{placeholder}` invokes getter 
	 * 	`placeholder` from the suite object.
	 *
	 * @param {Suite} suite
	 * @return {void}
	 */
	suite(suite) {
		this.separator();
		Object.keys(SUITE).forEach(message => {
			let options = SUITE[message];
			let suiteVars = message.match(/({.+?})/g) || [];
			suiteVars.forEach(placeholder => {
				let value;
				let prop = placeholder.replace("{", "").replace("}", "");
				if (typeof suite[prop] === "function") {
					value = suite[prop]();
				}
				else value = suite[prop];
				message = message.replace(placeholder, value);
			});
			this.log(message, options);
		});
	}


	/**
	 * Generate a separator string.
	 *
	 * Length of string is minimum of console width or 48 characters.
	 *
	 * @param {string} char optional Character to use (`-`)
	 * @param {bool} newline optional Extra newline after output (`false`)
	 * @return {void}
	 */
	separator(char="-", newline=false) {
		let repeat = Math.min(process.stdout.columns, 48);
		this.log(`\n ${char.repeat(repeat)}`, {bright: true});
		if (newline) this.log("");
	}


	/**
	 * Get an option from an options object.
	 *
	 * Returns the value in the object if set, or `optionVal`.
	 *
	 * @param {object} options Options object
	 * @param {string} option Option name to get value of
	 * @param {mixed} optionVal Default to return if not set
	 */
	option(options, option, optionVal=null) {
		return options.hasOwnProperty(option) ? options[option] : optionVal;
	}


	/**
	 * Prepare options.
	 *
	 * @param {object} options
	 * @return {string}
	 */
	options(options={}) {
		let str = "";
		Object.keys(ANSI).forEach(option => {
			if (ANSI_BOOLEAN.includes(option) && options[option]) {
				str += ANSI[option];
			}
			else if (options[option]) {
				str += ANSI[option][options[option]];
			}
		});
		return str;
	}


	/**
	 * Indentation string
	 *
	 * @param {int} indentation optional
	 * @return {string}
	 */
	indentation(indentation=this.indent) {
		if (indentation < 0) indentation = 0;
		return "\t".repeat(indentation * INDENTATION);
	}


	/**
	 * Get a message for stdout.
	 *
	 * Applies formatting, indentation, resets and newlines as required.
	 *
	 * Options:
	 * - `newline` - Add `\n` after message
	 * - `indent` - Indentation level override
	 * - `reset` - Reset formatting before output
	 * - `resetAfter` - Reset formmatting after output
	 *
	 * @param {string} message
	 * @param {object} options Options object
	 * @return {string}
	 */
	stdout(message, options={}) {
		let stdout = "";
		if (this.option(options, "newline", false)) {
			message += "\n";
		}
		if (this.option(options, "reset", true)) {
			stdout += ANSI.reset;
		}
		stdout += this.indentation(this.option(options, "indent"));
		stdout += this.options(options);
		stdout += message;
		if (this.option(options, "resetAfter", true)) {
			stdout += ANSI.reset;
		}
		return stdout;
	}

}

export default Logger;