import Timer from "./Timer.mjs";

/**
 * Suite
 *
 * @package Tramline
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class Suite {

	/**
	 * Constructor.
	 *
	 * @return {self}
	 */
	constructor() {

		/**
		 * Tests
		 *
		 * @var {array}
		 */
		this.tests = [];

		/**
		 * Timer
		 *
		 * @var {Timer}
		 */
		this.Timer = new Timer();

	}


	/**
	 * End.
	 *
	 * Ends the timer.
	 *
	 * @return {void}
	 */
	end() {
		this.Timer.end();
	}


	/**
	 * Add a test.
	 *
	 * @param {Test} test
	 * @return {void}
	 * @throws {Error} If test not completed run.
	 */
	add(test) {
		if (!test.run) {
			throw new Error("Test not run.");
		}
		this.tests.push(test);
	}


	/**
	 * Get count of passed tests.
	 *
	 * @return {int}
	 */
	get passed() {
		return this.tests.filter(t => t.passed).length;
	}


	/**
	 * Get count of failed tests.
	 *
	 * @return {int}
	 */
	get failed() {
		return this.tests.filter(t => !t.passed).length;
	}


	/**
	 * Get count of all run tests.
	 *
	 * @return {int}
	 */
	get count() {
		return this.tests.length;
	}


	/**
	 * Get count of all run unique units.
	 *
	 * @return {int}
	 */
	get countUnits() {
		let units = [];
		this.tests.forEach(test => {
			if (!units.includes(test.unit)) {
				units.push(test.unit);
			}
		});
		return units.length;
	}


	/**
	 * Get elapsed time from timer.
	 *
	 * @return {int} ms
	 */
	get timer() {
		return this.Timer.elapsed(true)
	}


	/**
	 * Get timer time formatted as a string to a given precision.
	 *
	 * @param {int} precision
	 * @return {string} s
	 */
	timerFormatted(precision=3) {
		return (this.timer / 1000).toFixed(precision);
	}


	/**
	 * Get test suite success status as a percentage.
	 *
	 * @return {float}
	 */
	get success() {
		let total = (this.passed + this.failed);
		if (total === 0) return 100;
		else return ((this.passed / total) * 100);
	}


	/**
	 * Get success percentage formatted as a string to a given precision.
	 *
	 * @param {int} precision
	 * @return {string}
	 */
	successFormatted(precision=2) {
		return this.success.toFixed(precision);
	}


	/**
	 * Get whether all tests in the suite succeeded.
	 *
	 * @return {bool}
	 */
	get succeeded() {
		for (let i = 0; i < this.tests.length; i++) {
			if (!this.tests[i].passed) return false;
		}
		return true;
	}

}

export default Suite;