const TRAMLINE = "0.2.1";
import Cli from "./Cli.mjs";
import Config from "./Config.mjs";
import Loader from "./Loader.mjs";
import Logger from "./Logger.mjs";
import Test from "./Test.mjs";
import Unit from "./Unit.mjs";
import Suite from "./Suite.mjs";
import TestHookError from "./Errors/TestHookError.mjs";
import {assert} from "./Assertions.mjs";
import * as assertions from "./Assertions.mjs";

/**
 * Test hooks where failure is fatal.
 *
 * A failure in these hooks should abort the unit.
 *
 * @var {array}
 */
const TEST_HOOKS_FATAL = [
	"__init", "__run"
];

/**
 * Tramline
 *
 * @package Tramline
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class Tramline {

	/**
	 * Constructor.
	 *
	 * @param {bool} init optional Auto-init (`true`)
	 * @param {Config} config optional Configuration for auto-init
	 * @return {self}
	 */
	constructor(init=true, config=null) {

		/**
		 * Tests
		 *
		 * @var {array}
		 */
		this.tests = [];

		/**
		 * Config
		 *
		 * @var {Config}
		 */
		this.Config = null;

		/**
		 * Loader
		 *
		 * @var {Loader}
		 */
		this.Loader = null;

		/**
		 * Logger
		 *
		 * @var {Logger}
		 */
		this.Logger = null;

		/**
		 * Suite
		 *
		 * @var {Suite}
		 */
		this.Suite = null;


		/**
		 * Logger
		 * 
		 * @var {Logger}
		 */
		this.Logger = new Logger();


		/**
		 * Init
		 */
		if (init) this.init(config);

	}


	/**
	 * Construct.
	 *
	 * @param {Config} optional Config, or uses CLI config.
	 *
	 * @return {void}
	 */
	init(config=null) {

		/**
		 * Config
		 */
		this.Config = config ? new Config(config) : Cli.config();
		if (!this.Config.root) {
			this.exit(1, "(!) Exiting; no directory specified.");
		}

		/**
		 * Loader
		 */
		this.Loader = new Loader(this.Config);
		this.tests = this.Loader.load();

		/**
		 * App log
		 */
		this.Logger.app(Date.now(), this.Config, this.tests, TRAMLINE);

	}


	/**
	 * Run all currently loaded tests.
	 *
	 * @param {bool} log optional Log summary after completion (`true`)
	 * @param {bool} exit optional Exit after completion (`true`)
	 * @param {bool} catchErrors optional Catch errors (`true`)
	 * @return {Promise}
	 */
	run(log=true, exit=true, catchErrors=true) {
		//Imports
		let imports = [];

		//Create new suite
		this.Suite = new Suite();

		//Import each test unit
		this.tests.forEach(test => {
			imports.push(import(test.file));
		});

		//Error
		let e = (e) => {
			if (!catchErrors) throw e;
			this.Suite.end();
			this.Logger.log("");
			this.Logger.error(e);
			if (exit) {
				this.exit(1, "(!) Aborting; could not load all tests.");
			}
		}

		/**
		 * Promisify imports + run
		 */
		return new Promise((resolve, reject) => {

			/**
			 * Imports completed; run the units
			 */
			Promise.all(imports).then(imports => {

				/**
				 * Run the units
				 */
				imports.forEach(module => {
					this.unit(module.default, catchErrors);
				});

				resolve();

			}).then(() => {

				/**
				 * Complete suite
				 */
				this.Suite.end();
				if (log) this.Logger.suite(this.Suite);
				if (exit) this.exit(this.Suite.succeeded);

			}).catch(e);

		}).catch(e);
	}


	/**
	 * Run a unit.
	 *
	 * @param {Unit} unit Test unit definition
	 * @param {bool} catchErrors optional Catch errors (`true`)
	 * @return {void}
	 * @throws {Error} If unit fails and error catching disabled.
	 */
	unit(unit, catchErrors=true) {
		//Logger
		this.Logger.unit(unit);

		//Run the unit
		try {
			let instance = new unit();
			this.unitTest(instance);
		}
		catch (error) {
			if (!catchErrors) throw error;
			this.Logger.error(error);
			this.Logger.error(`Test unit "${unit.name}" failed.`);
		}
		finally {
			this.Logger.unitEnd();
		}
	}


	/**
	 * Run tests in a unit.
	 *
	 * Hooks:
	 * - `__init` always runs first.
	 * - Test methods are preceded by `__run` and succeeded by `__after`.
	 * - `__finally` always runs last.
	 *
	 * Arguments:
	 * - `__init` always gets nothing.
	 * - `__run` always gets return from `__init`.
	 * - Test methods get the return from `__run`, or `__init` if 
	 * 	`__run` is not defined.
	 * 	- `__finally` always gets return from `__init`.
	 *
	 * Results are also saved by `test(...)` to `results` in the unit.
	 *
	 * @param {Unit} unit
	 * @return {void}
	 */
	unitTest(unit) {
		let init, initArgs;
		[init, initArgs] = this.hook(unit, "__init");
		unit.__tests().forEach(test => {
			let run, runArgs, result;
			[run, runArgs] = this.hook(unit, "__run", initArgs);
			result = this.test(unit, test, (run ? runArgs : initArgs));
			this.hook(unit, "__after", result);
		});
		this.hook(unit, "__finally", initArgs);
	}


	/**
	 * Run a test method in a unit.
	 *
	 * @param {Unit} unit Test unit instance
	 * @param {string} method Test method name
	 * @param {array} args optional Arguments to pass to method
	 * @return {mixed} Return value from test method
	 */
	test(unit, method, args=[]) {
		//Create the test
		let result = null;
		let test = new Test(unit.constructor.name, method);
		
		//Process test args
		args = this.testArgs(args);

		//Run the test method
		try {
			result = unit[method](...args);
			test.passed = true;
		}
		catch (error) {
			test.error = error;
			test.passed = false;
		}
		finally {
			test.run = true;
			unit.results[method] = result;
		}

		//Only log test hooks if they failed
		if (!this.isTestHook(test) || !test.passed) {
			this.Suite.add(test);
			this.Logger.test(test);
		}
		//If a critical test hook failed, throw to abort unit
		if (!test.passed && this.isTestHookFatal(test)) {
			throw new TestHookError(test.unit, test.title);
		}
		return result;
	}


	/**
	 * Process test method arguments by converting to array for unpacking.
	 *
	 * @param {mixed} args
	 * @return {array}
	 */
	testArgs(args) {
		if (!Array.isArray(args)) {
			args = [args];
		}
		return args;
	}


	/**
	 * Run a test hook in a unit.
	 *
	 * @param {Unit} unit Unit instance
	 * @param {string} hook Hook name
	 * @param {array} args optional Arguments for hook
	 * @return {array} 0 -> bool ran, 1 -> return value if ran
	 */
	hook(unit, hook, args=[]) {
		if (typeof unit[hook] === "function") {
			return [true, this.test(unit, hook, args)];
		}
		else return [false, null];
	}


	/**
	 * Get whether a test object refers to a hook.
	 *
	 * Test hooks are meta-tests whch refer to suite/unit/test operation.
	 *
	 * Test hooks are generally executed as for regular tests but are 
	 * 	not added to logs or test suite statistics unless they fail.
	 *
	 * This reduces unnecessary verbosity due to e.g. the `__run` hook 
	 * 	being called with identical output for each test in a unit.
	 *
	 * This method determines whether a generic `Test` is a hook.
	 *
	 * **ANY** test (test method) that starts with `__` is a hook.
	 *
	 * @param {Test} test
	 * @return {bool}
	 */
	isTestHook(test) {
		return test.title.startsWith("__");
	}


	/**
	 * Get whether a test is a fatal test hook.
	 *
	 * Test hooks such as `__init`/`__run` (defined in the 
	 * 	`TEST_HOOKS_FATAL` constant) should fail the unit on fail.
	 *
	 * @param {Test} test
	 * @return {bool}
	 */
	isTestHookFatal(test) {
		return TEST_HOOKS_FATAL.includes(test.title);
	}


	/**
	 * Exit.
	 *
	 * @param {bool} optional ok OK for status code
	 * @param {string} msg optional Message to log
	 * @return {void}
	 */
	exit(ok=true, msg=null) {

		/**
		 * Log
		 */
		if (msg) this.Logger.log(msg);
		this.Logger.separator("=", true);

		/**
		 * Exit
		 */
		if (!ok) process.exit(1);
		else process.exit(0);

	}
	
}

export default Tramline;
export {Tramline, Unit, assert, assertions};