import Response from "../response.mjs";
import RequestError from "../errors/RequestError.mjs";

/**
 * Request XHR interface
 *
 * @package Requesty
 * @subpackage Interfaces
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class XhrInterface {

	/**
	 * Make a request.
	 *
	 * @param {Request} request Requesty request object
	 * @param {function} resolve Success callback
	 * @param {function} reject Error callback
	 * @return {XMLHttpRequest}
	 */
	request(request, resolve, reject) {
		let xhr = new XMLHttpRequest();
		xhr.open(request.method, request.url);
		xhr = this.attachXhrHeaders(xhr, request.headers);
		xhr = this.attachXhrCallbacks(xhr, resolve, reject);
		xhr.send(request.payload);
		return xhr;
	}


	/**
	 * Prepare a request response from an `XMLHttpRequest`.
	 *
	 * @param {Request} request
	 * @param {XMLHttpRequest} xhr response
	 * @return {Response}
	 */
	response(request, xhr) {
		let response = new Response();
		response.body = xhr.response;
		response.bodyRaw = xhr.responseText;
		response.type = xhr.responseType;
		response.status = xhr.status;
		response.message = xhr.statusText;
		response.request = request;
		response.interface = xhr;
		response.headers = this.responseHeaders(xhr);
		return response;
	}


	/**
	 * Get response headers object from XMLHttpRequest.
	 *
	 * @param {XMLHttpRequest} xhr
	 * @return {object}
	 */
	responseHeaders(xhr) {
		let headerMap = {};
		let headers = xhr.getAllResponseHeaders();
		let headersArray = headers.trim().split(/[\r\n]+/);
		headersArray.forEach(header => {
			let kv = header.split(":");
			headerMap[kv[0]] = (kv.length > 1) ? kv[1].trim() : "";
		});
		return headerMap;
	}


	/**
	 * Add headers to a request.
	 *
	 * @param {XMLHttpRequest} xhr
	 * @param {object} headers
	 * @return {XMLHttpRequest}
	 */
	attachXhrHeaders(xhr, headers) {
		Object.keys(headers).forEach(k => {
			xhr.setRequestHeader(k, headers[k]);
		});
		return xhr;
	}


	/**
	 * Attach callbacks to a request.
	 *
	 * @param {XMLHttpRequest} xhr
	 * @param {function} resolve Success callback
	 * @param {function} reject Error callback
	 * @return {XMLHttpRequest}
	 */
	attachXhrCallbacks(xhr, resolve, reject) {
		xhr = this.attachXhrOnLoad(xhr, resolve, reject);
		xhr = this.attachXhrOnError(xhr, resolve, reject);
		return xhr;
	}


	/**
	 * Attach XHR load handler.
	 *
	 * @param {XMLHttpRequest} xhr
	 * @param {function} resolve Success callback
	 * @param {function} reject Error callback
	 * @return {XMLHttpRequest}
	 */
	attachXhrOnLoad(xhr, resolve, reject) {
		xhr.onload = function() {
			if (this.statusSuccess(xhr.status)) {
				resolve(xhr);
			}
			else {
				reject(this.xhrRequestError(xhr));
			}
		}.bind(this);
		return xhr;
	}


	/**
	 * Attach XHR error handler.
	 *
	 * @param {XMLHttpRequest} xhr
	 * @param {function} resolve Success callback
	 * @param {function} reject Error callback
	 * @return {XMLHttpRequest}
	 */
	attachXhrOnError(xhr, resolve, reject) {
		xhr.onerror = function() {
			reject(this.xhrRequestError(xhr));
		}.bind(this);
		return xhr;
	}


	/**
	 * Create request error from an `XMLHttpRequest`.
	 * 
	 * Sets status code/status message according to 
	 *  the current request properties.
	 * 
	 * @param {XMLHttpRequest} xhr
	 * @return {RequestError}
	 */
	xhrRequestError(xhr) {
		return new RequestError(xhr.status, xhr.statusText);
	}


	/**
	 * Get whether request succeeded from status code.
	 *
	 * Status codes 200-299 are considered successful.
	 *
	 * @param {int} status
	 * @return {bool}
	 */
	statusSuccess(status) {
		return (status >= 200 && status < 300);
	}

}

export default new XhrInterface();