/**
 * Request maker
 *
 * @package Requesty
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class Maker {

	/**
	 * Constructor.
	 *
	 * @return {self}
	 */
	constructor() {

		/**
		 * Allow mocking `File` for tests
		 *
		 * @type {File}
		 */
		this.File = null;

		/**
		 * Allow mocking `FileList` for tests
		 *
		 * @type {FileList}
		 */
		this.FileList = null;

		/**
		 * Allow mocking `FormData` for tests
		 *
		 * @type {FormData}
		 */
		this.FormData = null;

		/**
		 * Get object references
		 */
		if (typeof File !== "undefined") this.File = File;
		if (typeof FileList !== "undefined") this.FileList = FileList;
		if (typeof FormData !== "undefined") this.FormData = FormData;
		
	}


	/**
	 * Make a request.
	 *
	 * @param {Request} request Requesty request
	 * @param {object} config Config object
	 * @return {Request}
	 */
	make(request, config) {
		if (this.urlNeedsBase(request.url) && config.urlBaseEnabled) {
			request.url = config.urlBase + request.url;
		}
		request = this.makeBody(request, config);
		request = this.makeHeaders(request, config);
		return request;
	}


	/**
	 * Make request body for method.
	 *
	 * @param {Request} request Requesty request
	 * @param {object} config Config object
	 * @return {Request}
	 */
	makeBody(request, config) {
		let method;
		if (request.method === "GET") {
			method = "makeBodyGET";
		}
		else method = "makeBodyPOST";
		return this[method](request);
	}


	/**
	 * Make the body for a GET request.
	 *
	 * Appends body values to the URL as an encoded string, if the 
	 * 	body is an object, or appends it directly to the URL otherwise.
	 *
	 * @param {Request} request Requesty request
	 * @return {Request}
	 */
	makeBodyGET(request) {
		request.url += this.urlEncode(request.body);
		return request;
	}


	/**
	 * Make the body for a POST request.
	 *
	 * If no content type header is specified, assumes form data with 
	 * 	`multipart/form-data` and will build a `FormData` object from the 
	 * 	request body if the body isn't one already. If the body is not 
	 * 	an instance of `Object`, the body content will be added to a new 
	 * 	`FormData` object under the `formdata` key.
	 *
	 * If content type is specified, attempts to follow it. If the content 
	 * 	type is JSON, automatically JSON encodes the body unless it's 
	 * 	a string already. If the content type is URL-encoded form, URL 
	 * 	encodes it automatically if the body is an instanceof `Object`, or 
	 * 	otherwise appends the body to the URL directly.
	 *
	 * Finally, if the content type is explicitly set to form data, 
	 * 	automatically creates a `FormData` object if required in the 
	 * 	same manner as described above for implicit form data submission.
	 *
	 * @param {Request} request Requesty request
	 * @return {Request}
	 */
	makeBodyPOST(request) {
		let ct = request.getHeader("Content-Type");
		if (ct === "application/json") {
			if (typeof request.body !== "string") {
				request.payload = JSON.stringify(request.body);
			}
			else request.payload = request.body;
		}
		else if (ct === "multipart/form-data" || !ct) {
			if (this.FormData && !(request.body instanceof this.FormData)) {
				request.payload = this.makeFormData(request.body);
			}
			else request.payload = request.body;
		}
		else if (ct === "application/x-www-form-urlencoded") {
			request.payload = this.urlEncode(request.body);
		}
		else request.payload = request.body;
		return request;
	}


	/**
	 * Make request headers.
	 *
	 * Applies default headers.
	 *
	 * @param {Request} request Requesty request
	 * @param {object} config Config object
	 * @return {Request}
	 */
	makeHeaders(request, config) {
		if (config.defaultHeadersEnabled) {
			Object.assign(request.headers, config.defaultHeaders);
		}
		return request;
	}


	/**
	 * Make a form data object from an object.
	 *
	 * Enumerates the keys in `obj` and assigns them to a `FormData`.
	 *
	 * If passed input which is not an object, adds it to a new `FormData` 
	 * 	instance under the `formdata` key.
	 *
	 * @param {object} obj
	 * @return {FormData}
	 */
	makeFormData(obj) {
		let fd = new this.FormData();
		if (obj instanceof Object) {
			for (let key in obj) {
				if (obj[key] instanceof this.File) {
					fd.append(`${key}[]`, obj[key]);
				}
				else if (obj[key] instanceof this.FileList) {
					for (let i = 0; i < obj[key].length; i++) {
						fd.append(`${key}[]`, obj[key][i]);
					}
				}
				else {
					fd.append(key, this.makeStringValue(obj[key]));
				}
			}
		}
		else fd.append("formdata", obj);
		return fd;
	}


	/**
	 * Make a value into a string for use in e.g. an request body.
	 *
	 * If is an instance of `Object`, represents as JSON.
	 *
	 * @param {mixed} value
	 * @return {mixed}
	 */
	makeStringValue(value) {
		if (value instanceof Object) {
			return JSON.stringify(value);
		}
		else return value;
	}


	/**
	 * Convert an object to encoded URL.
	 *
	 * When not passed an object, the input parameter is returned unchanged.
	 *
	 * @param {object} obj
	 * @return {string} Starts with `?`
	 */
	urlEncode(obj) {
		let i = 0;
		let url = "";
		if (obj instanceof Object) {
			for (let key in obj) {
				if (i === 0) {
					i++;
					url += "?";
				}
				else url += "&";

				url += `${key}=${this.makeStringValue(obj[key])}`;
			}
			return url;
		}
		else return obj;
	}


	/**
	 * Get whether a URL needs a base prepending or already includes one.
	 *
	 * Looks for a protocol root to determine if the URL is qualified.
	 *
	 * @return {bool}
	 */
	urlNeedsBase(url) {
		return (url.indexOf("://") === -1);
	}

}

export default new Maker();