import BodyKeyExistsError from "./errors/BodyKeyExistsError.mjs";
import HeaderExistsError from "./errors/HeaderExistsError.mjs";

/**
 * Request class
 *
 * Properties:
 * - url - URL to call; auto-prepended with root base (if enabled)
 * - method - Method to call with ("GET")
 * - body - Body data object to send
 * - headers - Headers object to apply (merged with default headers)
 * - interface - Internal property to identify interface to use ("xhr")
 * - payload - Internal; processed payload to send (do not alter directly)
 * 
 * @package Requesty
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class Request {

	/**
	 * Construct request.
	 *
	 * @param {string} url URL, will be prepended with base
	 * @param {string} method optional Method to use ("GET")
	 * @return {self}
	 */
	constructor(url, method="GET") {
		this.url = url;
		this.method = method;
		this.headers = {};
		this.body = {};
		this.interface = "xhr";
		this.payload = null;
	}


	/**
	 * Add a new body key if not already set.
	 *
	 * @param {string} key Key
	 * @param {string} value Value
	 * @return {void}
	 * @throws {BodyKeyExistsError} If key already set.
	 */
	addBodyKey(key, value) {
		if (!this.bodyKeySet(key)) {
			this.setBodyKey(key, value);
		}
		else throw new BodyKeyExistsError(key);
	}


	/**
	 * Set body key even if already set.
	 *
	 * @param {string} key Key
	 * @param {string} value Value
	 * @return {void}
	 */
	setBodyKey(key, value) {
		this.body[key] = value;
	}


	/**
	 * Get body key value.
	 *
	 * @param {string} key
	 * @return {string|null}
	 */
	getBodyValue(key) {
		return this.body[key];
	}


	/**
	 * Get if body key is set.
	 *
	 * @param {string} key
	 * @return {bool}
	 */
	bodyKeySet(key) {
		return this.body.hasOwnProperty(key);
	}


	/**
	 * Add a header.
	 *
	 * @param {string} header Header
	 * @param {string} value Value
	 * @return {void}
	 * @throws {HeaderExistsError} If header already exists.
	 */
	addHeader(header, value) {
		if (!this.headerSet(header)) {
			this.setHeader(header, value)
		}
		else throw new HeaderExistsError(header);
	}


	/**
	 * Set header value (even if already exists).
	 *
	 * @param {string} header Header
	 * @param {string} value Value
	 * @return {void}
	 */
	setHeader(header, value) {
		this.headers[header] = value;
	}


	/**
	 * Get header value.
	 *
	 * @param {string} header Header
	 * @return {string|null}
	 */
	getHeader(header) {
		return this.headers[header];
	}


	/**
	 * Get whether a header is set.
	 *
	 * @param {string} header Header
	 * @return {bool}
	 */
	headerSet(header) {
		return this.headers.hasOwnProperty(header);
	}

}

export default Request;