import config from "./config.mjs";
import Maker from "./maker.mjs";
import Request from "./request.mjs";
import RequestInterfaces from "./interfaces.mjs";
import RequestyError from "./errors/RequestyError.mjs";

/**
 * Requesty
 *
 * @package Requesty
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class Requesty {

	/**
	 * Constructor.
	 *
	 * @return {self}
	 */
	constructor() {

		/**
		 * Requesty version
		 *
		 * @var {string}
		 */
		this.requesty = "1.4.2";

	}


	/**
	 * Endpoint to create and send a request and get a response 
	 * 	without manually creating a Request object.
	 *
	 * To specify request properties, such as method, body and headers, 
	 * 	define them in the `request` options object – this object is 
	 * 	assigned directly to the request object, so set properties as 
	 * 	for the `Request` class.
	 *
	 * @param {string} url
	 * @param {object} request optional Properties to add to request object
	 * @param {object} config optional Override default configuration
	 * @return {Promise} Promise resolved with response when request complete
	 * @throws {RequestyError} Request object not valid or request fails.
	 */
	req(url, request={}, config=null) {
		let req = new Request(url);
		Object.assign(req, request);
		return this.request(req, config);
	}
	

	/**
	 * Make a request.
	 *
	 * @param {Request} request Requesty request
	 * @param {object} conf optional Override default configuration
	 * @return {Promise} Promise resolved with response when request complete
	 * @throws {RequestyError} Request object not valid or request fails.
	 */
	request(request, conf=null) {
		let requestInterface;
		if (!conf) conf = config;
		if (!(request instanceof Request)) {
			throw new RequestyError("Request not a Requesty request.");
		}

		request = Maker.make(request, conf);
		requestInterface = RequestInterfaces[request.interface];
		return new Promise(function(resolve, reject) {
			return requestInterface.request(request, resolve, reject);
		}).then(function(response) {
			return requestInterface.response(request, response);
		});
	}

}

let requesty = new Requesty();
export default requesty;
export {requesty as main, config, Request};