import Unit from "tramline/src/Unit";
import * as tramline from "tramline";
import Config from "../src/config.mjs";
import HeaderExistsError from "../src/errors/HeaderExistsError.mjs";

/**
 * Assertions
 *
 * @var {object}
 */
const assertions = tramline.assertions;

/**
 * Header
 *
 * @var {string}
 */
const HEADER = "foo";

/**
 * Header value
 *
 * @var {string}
 */
const HEADER_VALUE = "bar"


/**
 * Config test
 *
 * @package Requesty
 * @subpackage Tests
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class ConfigTest extends Unit {

	/**
	 * Get new config for every test.
	 *
	 * @return {Config}
	 */
	__run() {
		return new Config.constructor();
	}


	/**
	 * Construction test.
	 *
	 * @param {Config} config
	 * @return {void}
	 */
	construction(config) {
		assertions.assert(config.urlBase.constructor === String);
		assertions.assert(config.hasOwnProperty("urlBaseEnabled"));
		assertions.assert(config.hasOwnProperty("defaultHeaders"));
		assertions.assert(config.hasOwnProperty("defaultHeadersEnabled"));
	}


	/**
	 * Add headers.
	 *
	 * @param {Config} config
	 * @return {Config} Can be used for other tests
	 */
	addHeader(config) {
		config.addHeader(HEADER, HEADER_VALUE);
		assertions.assert(config.defaultHeaders.hasOwnProperty(HEADER));
		assertions.identical(config.defaultHeaders[HEADER], HEADER_VALUE);
		return config;
	}


	/**
	 * Add header of same name as existing (should throw).
	 *
	 * @param {Config} config
	 * @return {void}
	 */
	addHeaderExisting(config) {
		config = this.addHeader(config);
		assertions.throws(() => {
			config.addHeader(HEADER, "foobar");
		}, HeaderExistsError);
	}


	/**
	 * Set headers.
	 *
	 * @param {Config} config
	 * @return {Config} Can be used for other tests
	 */
	setHeader(config) {
		config.setHeader(HEADER, "foobar");
		assertions.assert(config.defaultHeaders.hasOwnProperty(HEADER));
		assertions.identical(config.defaultHeaders[HEADER], "foobar");
		return config;
	}


	/**
	 * Set an existing header.
	 *
	 * @param {Config} config
	 * @return {void}
	 */
	setHeaderExisting(config) {
		config = this.setHeader(config);
		config.setHeader(HEADER, "test");
		assertions.assert(config.defaultHeaders.hasOwnProperty(HEADER));
		assertions.identical(config.defaultHeaders[HEADER], "test");
	}


	/**
	 * Test getting headers.
	 *
	 * @param {Config} config
	 * @return {void}
	 */
	getHeader(config) {
		config = this.addHeader(config);
		assertions.identical(config.getHeader("123"), undefined);
		assertions.identical(config.getHeader(HEADER), HEADER_VALUE);
	}


	/**
	 * Test getting if headers are set.
	 *
	 * @param {Config} config
	 * @return {void}
	 */
	headerSet(config) {
		config = this.addHeader(config);
		assertions.truthy(config.headerSet(HEADER));
		assertions.falsey(config.headerSet("tramline"));
	}

}

export default ConfigTest;