import Unit from "tramline/src/Unit";
import * as tramline from "tramline";
import Maker from "../src/maker.mjs";

/**
 * Assertions
 *
 * @var {object}
 */
const assertions = tramline.assertions;


/**
 * File mock.
 *
 * @package Requesty
 * @subpackage Tests
 * @author James Walker
 * @copyright James Walker 2019
 * @license MIT License
 */
class File {

	/**
	 * Foobar
	 *
	 * @return {sting}
	 */
	get foobar() {
		return "foobar";
	}

}


/**
 * File list mock.
 *
 * @package Requesty
 * @subpackage Tests
 * @author James Walker
 * @copyright James Walker 2019
 * @license MIT License
 */
class FileList {}


/**
 * FormData mock.
 *
 * @package Requesty
 * @subpackage Tests
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class FormData {

	/**
	 * Constructor.
	 *
	 * @return {self}
	 */
	constructor() {

		/**
		 * Data
		 *
		 * @var {object}
		 */
		this.data = {};

	}


	/**
	 * Append data.
	 *
	 * When the key exists, uses `${key}1` instead (only two copies).
	 *
	 * @param {string} key
	 * @param {mixed} value
	 * @return {void}
	 */
	append(key, value) {
		if (this.data.hasOwnProperty(key)) {
			this.data[`${key}1`] = value;
		}
		else this.data[key] = value;
	}


	/**
	 * Get a key value.
	 *
	 * @param {string} key
	 * @return {void}
	 */
	get(key) {
		return this.data[key];
	}


	/**
	 * Get whether a key is set.
	 *
	 * @param {string} key
	 * @return {bool}
	 */
	has(key) {
		return this.data.hasOwnProperty(key);
	}

}


/**
 * Maker test.
 *
 * @package Requesty
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class MakerTest extends Unit {

	/**
	 * Get new Maker for each test.
	 *
	 * @return {Maker}
	 */
	__run() {
		let maker = new Maker.constructor();
		maker.FormData = FormData;
		maker.File = File;
		maker.FileList = FileList;
		return maker;
	}


	/**
	 * Construction.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	construction(maker) {
		assertions.assert(maker.hasOwnProperty("FormData"));
		assertions.assert(maker.hasOwnProperty("File"));
		assertions.assert(maker.hasOwnProperty("FileList"));
		assertions.assert(maker.FormData === FormData);
		assertions.assert(maker.File === File);
		assertions.assert(maker.FileList === FileList);
	}


	/**
	 * Test request making.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	make(maker) {
		let body = {foo: true, obj: {foobar: true}};
		let requests = [
			{
				request: {
					url: "/relative",
					body: body,
					method: "POST",
					headers: {Request: "Header"},
				},
				config: {
					urlBase: "http://localhost",
					urlBaseEnabled: true,
					defaultHeaders: {foo: "bar"},
					defaultHeadersEnabled: true
				}
			},
			{
				request: {
					url: "http://foobar/relative",
					body: body,
					method: "GET",
					headers: {}
				},
				config: {
					urlBase: "http://localhost",
					urlBaseEnabled: true,
					defaultHeaders: {foo: "bar"},
					defaultHeadersEnabled: true
				}
			},
			{
				request: {
					url: "/relative",
					body: body,
					method: "GET",
					headers: {}
				},
				config: {
					urlBase: "http://localhost",
					urlBaseEnabled: false,
					defaultHeaders: {foo: "bar"},
					defaultHeadersEnabled: false
				}
			}
		];

		/**
		 * Make each request
		 */
		requests.forEach(req => {
			let request, config;
			[request, config] = Object.values(req);
			if (request.method === "POST") request.getHeader = () => null;
			let made = maker.make(Object.assign({}, request), config);

			/**
			 * URL
			 */
			let eUrl = request.url;
			if (config.urlBaseEnabled && maker.urlNeedsBase(request.url)) {
				eUrl = config.urlBase + request.url;
			}
			//Cannot use `identical` - GET URLs have body...
			assertions.assert(made.url.indexOf(eUrl) === 0);

			/**
			 * Headers
			 */
			let all = Object.assign({}, request.headers);
			if (config.defaultHeadersEnabled) {
				all = Object.assign(all, config.defaultHeaders);
			}
			Object.keys(all).forEach(header => {
				assertions.assert(made.headers.hasOwnProperty(header));
				assertions.identical(made.headers[header], all[header]);
			});

			/**
			 * Body
			 */
			let m = request.method;
			if (request.method !== "GET") {
				assertions.assert(made.payload instanceof FormData);
				Object.keys(request.body).forEach(b => {
					let ex = maker.makeStringValue(request.body[b]);
					assertions.assert(made.payload.has(b));
					assertions.identical(made.payload.get(b), ex);
				});
			}
			else {
				assertions.assert(made.payload == null);
				Object.keys(request.body).forEach(b => {
					let prep = maker.makeStringValue(request.body[b]);
					let expected = `${b}=${prep}`;
					assertions.assert(made.url.indexOf(expected) !== -1);
				})
			}
		});
	}


	/**
	 * Test `makeBody(...)` works properly for GET/other methods.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeBody(maker) {
		let body = {foo: "bar", obj: {foobar: true}};
		let requestGET = {method: "GET", body: body};
		let requestPUT = {method: "PUT", body: body};
		requestPUT.getHeader = () => null;
		let madeGET = maker.makeBody(requestGET, {});
		let madePUT = maker.makeBody(requestPUT, {});
		assertions.assert(madeGET.payload == null);
		assertions.assert(madePUT.payload instanceof FormData);
	}


	/**
	 * Test production of GET body.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeBodyGet(maker) {
		let i = 0;
		let body = {
			foo: "bar",
			sublime: true,
			obj: {foobar: true}
		};
		let request = {body: body};
		let made = maker.makeBodyGET(request);
		Object.keys(body).forEach(b => {
			i++;
			let prep = maker.makeStringValue(body[b]);
			let expected = `${i === 1 ? "?" : "&"}${b}=${prep}`;
			assertions.assert(made.url.indexOf(expected) !== -1)
		});
	}


	/**
	 * Test production of POST body - basic return type.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeBodyPOST(maker) {
		let request = {body: new FormData()};
		request.getHeader = () => null;
		assertions.identical(maker.makeBodyPOST(request), request);
	}


	/**
	 * Making body with no content type header uses `FormData`.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeBodyPOSTNoContentType(maker) {
		let request = {body: {foo: "bar", test: "true"}};
		request.getHeader = () => null;
		let result = maker.makeBodyPOST(request);
		assertions.assert(result.payload instanceof FormData);
		assertions.identical(result.payload.get("foo"), "bar");
		assertions.identical(result.payload.get("test"), "true");

		let request2 = {body: "foobar"};
		request2.getHeader = () => null;
		let result2 = maker.makeBodyPOST(request2);
		assertions.assert(result2.payload instanceof FormData);
		assertions.identical(result2.payload.get("formdata"), "foobar");

		let request3 = {body: new FormData()};
		request3.getHeader = () => null;
		request3.body.data.foo = "bar";
		let result3 = maker.makeBodyPOST(request3);
		assertions.identical(result3.payload, request3.body);
		assertions.identical(result3.payload.get("foo"), "bar");
	}


	/**
	 * Making body as `FormData`.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeBodyPOSTFormData(maker) {
		let request = {body: {foo: "bar", test: "true"}};
		request.getHeader = () => "multipart/form-data";
		let result = maker.makeBodyPOST(request);
		assertions.assert(result.payload instanceof FormData);
		assertions.identical(result.payload.get("foo"), "bar");
		assertions.identical(result.payload.get("test"), "true");

		let request2 = {body: "foobar"};
		request2.getHeader = () => "multipart/form-data";
		let result2 = maker.makeBodyPOST(request2);
		assertions.assert(result2.payload instanceof FormData);
		assertions.identical(result2.payload.get("formdata"), "foobar");
	}


	/**
	 * Making body as `FormData` when already is a `FormData` object.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeBodyPOSTFormDataExisting(maker) {
		let request = {body: new FormData()};
		request.getHeader = () => "multipart/form-data";
		request.body.data.foo = "bar";
		let result = maker.makeBodyPOST(request);
		assertions.identical(result.payload, request.body);
		assertions.identical(result.payload.get("foo"), "bar");
	}


	/**
	 * Making body with JSON content-type.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeBodyPOSTJson(maker) {
		let request = {body: {
			foo: "bar",
			test: "true",
		}};
		request.getHeader = () => "application/json";
		assertions.identical(
			maker.makeBodyPOST(request).payload,
			`{"foo":"bar","test":"true"}`
		);
	}


	/**
	 * Making body with JSON content-type and body already is string.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeBodyPOSTJsonAlreadySet(maker) {
		let request = {body: `{"foo": "bar", "test": "true"}`};
		request.getHeader = () => "application/json";
		assertions.identical(
			maker.makeBodyPOST(request).payload,
			`{"foo": "bar", "test": "true"}`
		);
	}


	/**
	 * Making body as URL-encoded string.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeBodyPOSTUrlEncoded(maker) {
		let request = {body: {foo: "bar", "test": "true"}};
		request.getHeader = () => "application/x-www-form-urlencoded";
		assertions.identical(
			maker.makeBodyPOST(request).payload,
			`?foo=bar&test=true`
		);
	}


	/**
	 * Making body with a custom content-type attaches data as-is.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeBodyPOSTCustomCT(maker) {
		let requests = [
			{body: "foobar 123456"},
			{body: "0x123456789 fake binary"},
			{body: "<h1>hello world</h1>"}
		];
		requests[0].getHeader = () => "fake";
		requests[1].getHeader = () => "binary";
		requests[2].getHeader = () => "text/html";
		requests.forEach(request => {
			assertions.identical(
				maker.makeBodyPOST(request).payload,
				request.body
			);
		});
	}


	/**
	 * Make headers.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeHeaders(maker) {
		let req = {headers: {}};
		let headers = {foo: "bar", test: true};
		let config = {
			defaultHeadersEnabled: true,
			defaultHeaders: headers
		};
		let request = maker.makeHeaders(req, config);
		Object.keys(headers).forEach(header => {
			assertions.assert(request.headers.hasOwnProperty(header));
			assertions.identical(request.headers[header], headers[header]);
		});
	}


	/**
	 * Making form data from object.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeFormData(maker) {
		let obj = {
			int: 1,
			foo: "bar",
			test: true,
			arr: [1, "foo"],
			obj: {foobar: true}
		};
		let fd = maker.makeFormData(obj);
		assertions.assert(fd instanceof FormData);
		Object.keys(obj).forEach(k => {
			let str = maker.makeStringValue(obj[k]);
			assertions.assert(fd.has(k));
			assertions.identical(fd.get(k), str);
		});
	}


	/**
	 * Making form data from a scalar value.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeFormDataFromScalar(maker) {
		let fd = maker.makeFormData("foobar");
		assertions.assert(fd instanceof FormData);
		assertions.assert(fd.has("formdata"));
		assertions.identical(fd.get("formdata"), "foobar");
	}


	/**
	 * Making form data with files.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeFormDataFiles(maker) {
		let test = {foo: "bar", f: new File(), attach: new FileList()};
		let made = maker.makeFormData(test);
		assertions.identical(made.get("foo"), "bar");
		assertions.identical(made.get("f[]").foobar, "foobar");
		// assertions.identical(made.get("attach[]"), "foo");
		// assertions.identical(made.get("attach[]1"), "bar");
	}


	/**
	 * Make headers when making disabled by config.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeHeadersDefaultDisabled(maker) {
		let req = {headers: {}};
		let headers = {foo: "bar", test: true};
		let request = maker.makeHeaders(req, {});
		assertions.assert(typeof request.headers === "object");
		assertions.assert(Object.keys(request.headers).length === 0);
	}


	/**
	 * Test conversion to string value.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	makeStringValue(maker) {
		let values = [
			[1, 1],
			["foobar", "foobar"],
			[[1, "foo"], "[1,\"foo\"]"],
			[{foobar: true}, "{\"foobar\":true}"]
		];
		values.forEach(v => {
			assertions.identical(maker.makeStringValue(v[0]), v[1]);
		});
	}


	/**
	 * Encoding URL.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	urlEncode(maker) {
		let tests = [
			{
				test: {foo: "bar", test: true},
				expect: "?foo=bar&test=true"
			},
			{
				test: "?foo=bar&test=true",
				expect: "?foo=bar&test=true"
			},
			{
				test: ["foo", "bar"],
				expect: "?0=foo&1=bar"
			}
		];
		tests.forEach(test => {
			assertions.identical(maker.urlEncode(test.test), test.expect);
		})
	}


	/**
	 * Test `urlNeedsBase(...)`.
	 *
	 * @param {Maker} maker
	 * @return {void}
	 */
	urlNeedsBase(maker) {
		assertions.truthy(maker.urlNeedsBase("/relative"));
		assertions.falsey(maker.urlNeedsBase("http://absolute"));
	}

}

export default MakerTest;