import Unit from "tramline/src/Unit";
import * as tramline from "tramline";
import Request from "../src/request.mjs";
import BodyKeyExistsError from "../src/errors/BodyKeyExistsError.mjs";
import HeaderExistsError from "../src/errors/HeaderExistsError.mjs";

/**
 * Assertions
 *
 * @var {object}
 */
const assertions = tramline.assertions;

/**
 * URL
 *
 * @var {string}
 */
const URL = "localhost";

/**
 * Body
 *
 * @var {string}
 */
const BODY = "foo";

/**
 * Body value
 *
 * @var {string}
 */
const BODY_VALUE = "bar";

/**
 * Header
 *
 * @var {string}
 */
const HEADER = "foo";

/**
 * Header value
 *
 * @var {string}
 */
const HEADER_VALUE = "bar";


/**
 * Request test
 *
 * @package Requesty
 * @subpackage Tests
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class RequestTest extends Unit {

	/**
	 * Return new request for each case.
	 *
	 * @return {Request}
	 */
	__run() {
		return new Request(URL);
	}


	/**
	 * Constructor parameters.
	 *
	 * @return {void}
	 */
	construct() {
		let request = new Request("localhost", "POST");
		assertions.identical(request.url, "localhost");
		assertions.identical(request.method, "POST");
	}


	/**
	 * Construction of properties.
	 *
	 * @param {Request} request
	 * @return {void}
	 */
	construction(request) {
		assertions.identical(request.url, URL);
		assertions.identical(request.method, "GET");
		assertions.assert(request.hasOwnProperty("headers"));
		assertions.assert(request.hasOwnProperty("body"));
		assertions.assert(request.hasOwnProperty("interface"));
		assertions.assert(request.hasOwnProperty("payload"));
	}


	/**
	 * Add a body key.
	 *
	 * @param {Request} request
	 * @return {Request} Can be used for other tests.
	 */
	addBodyKey(request) {
		request.addBodyKey(BODY, BODY_VALUE);
		assertions.assert(request.body.hasOwnProperty(BODY));
		assertions.identical(request.body[BODY], BODY_VALUE);
		return request;
	}


	/**
	 * Add a body key of same name as existing - should throw.
	 *
	 * @param {Request} request
	 * @return {void}
	 */
	addBodyKeyExisting(request) {
		request = this.addBodyKey(request);
		assertions.throws(() => {
			request.addBodyKey(BODY, "foobar");
		}, BodyKeyExistsError);
	}


	/**
	 * Set a body key value.
	 *
	 * @param {Request} request
	 * @return {Request} Can be used for other tests.
	 */
	setBodyKey(request) {
		request.setBodyKey(BODY, BODY_VALUE);
		assertions.assert(request.body.hasOwnProperty(BODY));
		assertions.identical(request.body[BODY], BODY_VALUE);
		return request;
	}


	/**
	 * Set body key even if already set.
	 *
	 * @param {Request} request
	 * @return {void}
	 */
	setBodyKeyExisting(request) {
		request = this.setBodyKey(request);
		request.setBodyKey(BODY, "test");
		assertions.assert(request.body.hasOwnProperty(BODY));
		assertions.identical(request.body[BODY], "test");
	}


	/**
	 * Get a body key value.
	 *
	 * @param {Request} request
	 * @return {void}
	 */
	getBodyKeyValue(request) {
		request = this.addBodyKey(request);
		assertions.identical(request.getBodyValue("123"), undefined);
		assertions.identical(request.getBodyValue(BODY), BODY_VALUE);
	}


	/**
	 * Get whether a body key is set.
	 *
	 * @param {Request} request
	 * @return {void}
	 */
	bodyKeySet(request) {
		request = this.addBodyKey(request);
		assertions.truthy(request.bodyKeySet(BODY));
		assertions.falsey(request.bodyKeySet("foobar"));
	}


	/**
	 * Add a header.
	 *
	 * @param {Request} request
	 * @return {Request} Can be used for other tests
	 */
	addHeader(request) {
		request.addHeader(HEADER, HEADER_VALUE);
		assertions.assert(request.headers.hasOwnProperty(HEADER));
		assertions.identical(request.headers[HEADER], HEADER_VALUE);
		return request;
	}


	/**
	 * Add header of same name as existing (should throw).
	 *
	 * @param {Request} request
	 * @return {void}
	 */
	addHeaderExisting(request) {
		request = this.addHeader(request);
		assertions.throws(() => {
			request.addHeader(HEADER, "foobar");
		}, HeaderExistsError);
	}


	/**
	 * Set headers.
	 *
	 * @param {Request} request
	 * @return {Request} Can be used for other tests.
	 */
	setHeader(request) {
		request.setHeader(HEADER, HEADER_VALUE);
		assertions.assert(request.headers.hasOwnProperty(HEADER));
		assertions.identical(request.headers[HEADER], HEADER_VALUE);
		return request;
	}


	/**
	 * Set an existing header.
	 *
	 * @param {Request} request
	 * @return {void}
	 */
	setHeaderExisting(request) {
		request = this.setHeader(request);
		request.setHeader(HEADER, "test");
		assertions.assert(request.headers.hasOwnProperty(HEADER));
		assertions.identical(request.headers[HEADER], "test");
	}


	/**
	 * Test getting headers.
	 *
	 * @param {Request} request
	 * @return {void}
	 */
	getHeader(request) {
		request = this.addHeader(request);
		assertions.identical(request.getHeader("123"), undefined);
		assertions.identical(request.getHeader(HEADER), HEADER_VALUE);
	}


	/**
	 * Test getting if headers are set.
	 *
	 * @param {Request} request
	 * @return {void}
	 */
	headerSet(request) {
		request = this.addHeader(request);
		assertions.truthy(request.headerSet(HEADER));
		assertions.falsey(request.headerSet("tramline"));
	}

}

export default RequestTest;