import Unit from "tramline/src/Unit";
import * as tramline from "tramline";
import Response from "../src/response.mjs";
import RequestError from "../src/errors/RequestError.mjs";
import XhrInterface from "../src/interfaces/XhrInterface.mjs";

/**
 * Assertions
 *
 * @var {object}
 */
const assertions = tramline.assertions;

/**
 * XhrInterface test
 *
 * E2E tests for `request(...)` are currently omitted as cannot readily 
 * 	test `XMLHttpRequest(...)` due to ininavailability of API in Node.
 *
 * @package Requesty
 * @subpackage Tests
 * @author James Walker
 * @copyright James Walker 2018
 * @license MIT License
 */
class XhrTest extends Unit {

	/**
	 * Get new interface for each test.
	 *
	 * @return {XhrInterface}
	 */
	__run() {
		return new XhrInterface.constructor();
	}


	/**
	 * Response construction.
	 *
	 * @param {XhrInterface} xhr
	 * @return {void}
	 */
	response(xhr) {
		let r = "REQUEST";
		let h = {"Foo": "Bar", "Test": "true"};
		let x = {
			response: "foo",
			responseText: "foo",
			responseType: "html",
			status: 200,
			statusText: "GREAT",
			getAllResponseHeaders: function() {
				let s = "";
				Object.keys(h).forEach(i => s += `${i}: ${h[i]}\n`);
				return s;
			}
		};
		let m = {
			response: "body",
			responseText: "bodyRaw",
			responseType: "type",
			status: "status",
			statusText: "message"
		};

		/**
		 * Response
		 */
		let response = xhr.response(r, x);
		assertions.identical(response.request, r);
		assertions.assert(response instanceof Response);
		assertions.assert(typeof response.headers === "object");
		assertions.assert(typeof response.interface === "object");

		/**
		 * Properties
		 */
		Object.keys(x).forEach(y => {
			if (y === "getAllResponseHeaders") return;
			assertions.identical(response[m[y]], x[y]);
			assertions.assert(response.interface.hasOwnProperty(y));
		});

		/**
		 * Headers
		 */
		Object.keys(h).forEach(i => {
			assertions.assert(response.headers.hasOwnProperty(i));
			assertions.identical(response.headers[i], h[i]);
		});
	}


	/**
	 * Test response header processing.
	 *
	 * @param {XhrInterface} xhr
	 * @return {void}
	 */
	responseHeaders(xhr) {
		let h = {
			"Foo": "Bar",
			"Test": "true",
			"Sublime": "Text"
		}
		let x = {
			getAllResponseHeaders: function() {
				let s = "";
				Object.keys(h).forEach(i => {
					s += `${i}: ${h[i]}\n`
				});
				return s;
			}
		};
		let heads = xhr.responseHeaders(x);
		Object.keys(h).forEach(i => {
			assertions.assert(heads.hasOwnProperty(i));
			assertions.identical(heads[i], h[i]);
		});
	}


	/**
	 * Test header processing.
	 *
	 * @param {XhrInterface} xhr
	 * @return {void}
	 */
	attachXhrHeaders(xhr) {
		let x = {
			headers: {},
			setRequestHeader: function(k, v) {
				this.headers[k] = v;
			}
		};
		let headers = {foo: "bar", test: true};
		x = xhr.attachXhrHeaders(x, headers);
		Object.keys(headers).forEach(h => {
			assertions.assert(x.headers.hasOwnProperty(h));
			assertions.identical(x.headers[h], headers[h]);
		});
	}


	/**
	 * Attach XHR callbacks.
	 *
	 * @param {XhrInterface} xhr
	 * @return {void}
	 */
	attachXhrCallbacks(xhr) {
		let x = {};
		let res = () => {throw new Error("resolve");};
		let rej = () => {throw new Error("reject");};
		x = xhr.attachXhrCallbacks(x, res, rej);
		assertions.assert(typeof x.onload === "function");
		assertions.assert(typeof x.onerror === "function");
	}


	/**
	 * Attach XHR `onload`.
	 *
	 * @param {XhrInterface} xhr
	 * @return {void}
	 */
	attachXhrOnLoad(xhr) {
		let ok = () => {throw new Error("ok");};
		let fail = () => {throw new Error("fail");};
		let x = xhr.attachXhrOnLoad({}, ok, fail);
		x.status = 200;
		assertions.throws(x.onload, "ok");
		x.status = 400;
		assertions.throws(x.onload, "fail");
	}


	/**
	 * Attach XHR `onerror`.
	 *
	 * @param {XhrInterface} xhr
	 * @return {void}
	 */
	attachXhrOnError(xhr) {
		let f = () => {throw new Error("foobar");};
		xhr = xhr.attachXhrOnError({}, null, f);
		assertions.throws(xhr.onerror, "foobar");
	}


	/**
	 * Test returns `RequestError` properly.
	 *
	 * @param {XhrInterface} xhr
	 * @return {void}
	 */
	xhrRequestError(xhr) {
		let mock = {status: 999, statusText: "foobar"};
		let err = xhr.xhrRequestError(mock);
		assertions.assert(err instanceof RequestError);
		assertions.identical(err.status, mock.status);
		assertions.identical(err.message, mock.statusText);
	}


	/**
	 * Test "successful" status codes.
	 *
	 * @param {XhrInterface} xhr
	 * @return {void}
	 */
	statusSuccess(xhr) {
		assertions.truthy(xhr.statusSuccess(200));
		assertions.truthy(xhr.statusSuccess(250));
		assertions.truthy(xhr.statusSuccess(299));
		assertions.falsey(xhr.statusSuccess(199));
		assertions.falsey(xhr.statusSuccess(300));
	}

}

export default XhrTest;